(*--------------------------
The Original Perceptron
--------------------------

This article is based on a first implementation of the famous perceptron, using a 20x20 grid (just like the original Mark 1 Perceptron had). A perceptron is like the mother of an artificial neural network (ANN).
For the purposes of this tutorial I use the simple function y = 2x + 1 to restrict the memory, disk and cpu use. There are of course scripts with larger data- and trainingsets and larger embedding space that could give additional accuracy points.
We will see that rate of improvement drops quite markedly as you increase the number of training runs from 1 to 14.
A perceptron is an algorithm used in machine-learning. It it the simplest of all neural networks, consisting of only one neuron, and is typically used for pattern or image recognition.

The script will give an insight in the following 5 steps:
1. Define the target function to predict
2. Set an activator (sigmoid, linear)
3. Declare the loss function (error delta) to train
4. Declare an optimizer to minimize the loss (error)
5. Test and Predict a subset

First we start with the name of the program, which I hack in the editor of maXbox:
*)



program Perceptron1_31;

{$APPTYPE CONSOLE}

(*  #sign:max: MAXBOX8: 26/06/2018 20:31:11
 * implements a version of the algorithm set out at
 * http://natureofcode.com/book/chapter-10-neural-networks/ ,
 * but without graphics - ; classes assumed to be 1, -1
 * http://www.rosettacode.org/wiki/Perceptron#Pascal
 * 1_31: test calls with asserts
 *)


uses
  SysUtils;

type TAofReal= array[0..2] of real;
function feedForward(ins: array of integer; ws: array of real):
                                         integer; forward;
//We can change LEARNRATE to see another range output within an optimizer.

Const LEARNRATE = 0.01;
      BIAS = 1;



function targetOutput( a, b: integer ): integer;
//split in two classes: neg & pos against value b
begin
    if a * 2 + 1 < b then
      result:= 1
    else result:= -1;
end;

procedure showTargetOutput(var refval: string);
var x, y : integer;
begin
    for y:= 10 downto -9 do begin
        for x:= -9 to 10 do
           if targetOutput( x,y )= 1 then begin
                write( '#' );
                refval:= refval +'#'
           end else begin
                write( 'O' );
                refval:= refval +'O';
           end;
        writeln('')
    end;
    writeln('')
end;

procedure showOutput( ws: array of real; var resval: string );
var inputs: array[0..2] of integer;
    x, y: integer;
    tmpstr: string;
begin
    inputs[2]:= BIAS; (* bias *)
    tmpstr:=''; resval:='';
    for y:=10 downto-9 do begin
        for x:=-9 to 10 do begin
            inputs[0]:= x;
            inputs[1]:= y;
            if feedForward(inputs, ws) = 1 then
              tmpstr:= tmpstr+'#'
                else tmpstr:= tmpstr +'O';
        end;
        writeln(tmpstr);
        resval:= resval + tmpstr;
        tmpstr:='';
    end;
    writeln('');
end;



procedure randomWeights(var ws : TAofReal );
(* start with random weights -- pass by reference *)
var i: integer;
begin
    randomize; (* seed random-number generator *)
    for i:= 0 to 2 do
        ws[i]:= random * 2- 1;
end;

(*The objective is now to generate a feed forward code that allows to find the best parameters ws and b, that from input data, adjunct them to y data, in our case it will be a straight line defined by y_data = ws * x_data + b.
*)

function feedForward(ins: array of integer; ws: array of real): integer;
var sum: real; i : integer;
begin
    sum:= 0;
    for i:= 0 to 2 do
        sum:= sum+ ins[i] * ws[i];
    if sum > 0 then
         result:= 1
    else result:= -1
end;

(*In order for the perceptron to make the right decision, it needs to train with input for which the correct outcome is known, so that the weights can slowly be adjusted until they start producing the desired results. We can call this the
backpropagation call.
The emergent approach of an ANN resolves the adaptability and learning issues by building massively parallel models, analogous to neural networks, where information flow is represented by a (back) propagation of signals from the input
nodes. On the other hand,
emergent architectures are easier to design first, but they must be trained in order to produce useful behavior.

Of course there are plenty of optimizer routines like cross entropy or a derivative of the cross-entropy cost function for the softmax function. Cross entropy can be used to define the loss function in machine learning and optimization.
 There are many situations where cross-entropy needs
 to be measured but the distribution is unknown. An example is language modeling, where a model is created based on a training set, and then its cross-entropy is measured on a test set to assess how accurate the model is in predicting the test data.
In our perceptron we use a simple and single called delta routine which defines the error as the difference between target output and the feed forward calculation of the weights. By the way a cost function or a loss function means the same, namely to
reduce error difference and to gain accuracy.

     Error:= targetOutput(x, y) - feedForward(inputs, ws);

Entropy: This measure of disorder or impurity is based on the expected information content. Consider a message telling you about the class of a series of randomly drawn samples. The purer the set of samples, the more predictable this message
becomes (low entropy), and therefore the smaller
 the expected information.
Entropy is 0 if n samples in the data are the same.
Entropy is high if they are all different.
*)

procedure train(var ws: TAofReal; runs: integer; LEARNRATE: real);
(* pass the array of weights by reference so it can be modified *)
var inputs: array[0..2] of integer;
    error: real;
    x,y, i,j : integer;
begin
    inputs[2]:= BIAS; (* bias *)
    for i:= 1 to runs do begin
        for y:= 10 downto -9 do begin
           for x:= -9 to 10 do begin
               inputs[0]:= x;
               inputs[1]:= y;
               Error:= targetOutput(x, y )
                            - feedForward(inputs, ws);
             for j:= 0 to 2 do
                ws[j]:= ws[j]+ Error * inputs[j]* LEARNRATE;
           end;
        end;
    end;
end;



procedure testAll(const ws : TAofReal; runs : integer );
var inputs: array[0..2] of integer;
    x,y, i,j: integer;
    outputs: integer;
begin
    inputs[2]:= BIAS; (* bias *)
    for i:= 1 to runs do begin
        for y:= 10 downto -9 do begin
           for x:= -9 to 10 do begin
               inputs[0]:= x;
               inputs[1]:= y;
             for j:= 0 to 2 do begin
                outputs:= outputs+ round(ws[j]* inputs[j]);
             end;
            //print(itoa(outputs)+',  ')
            if outputs >= 1 then write('#') else write('O');
            outputs:= 0;
           end;
          writeln('')
        end;
    end;
end;

procedure Predict(const ws: TAofReal; a,b: integer);
var inputs: array[0..2] of integer;
    outputs, j: integer;
    cls: string;
begin
    inputs[2]:= BIAS; (* bias *)
    inputs[0]:= a;
    inputs[1]:= b;
    outputs:= 0;
    for j:= 0 to 2 do
       outputs:= outputs+ round(ws[j]* inputs[j]);
    if ((a*2)+1) < b then cls:=('pos') else cls:=('neg');
    WriteLn(inttostr(outputs)+',  is class: '+cls);
      {if cls = 'pos' then
           Assert((2*a)+1 < b,'Test Cond Error must <') else
           Assert((2*a)+1 >= b,'Test Cond Error must >=');}
end;

procedure PredictFloat(const ws: TAofReal; a,b: integer);
var inputs: array[0..2] of integer;
    j: integer; outp: Real;
    cls: string;
begin
    inputs[2]:= BIAS; (* bias *)
    inputs[0]:= a; inputs[1]:= b;
    outp:= 0;
    for j:= 0 to 2 do
       outp:= outp+ (ws[j]* inputs[j]);
    if ((a*2)+1) < b then cls:=('pos') else cls:=('neg');
    WriteLn(floattoStr(outp)+',  is class: '+cls);
       {if cls = 'pos' then
           Assert((2*a)+1 < b,'Test Cond Error must <') else
           Assert((2*a)+1 >= b,'Test Cond Error must >=');}
	  Readln;
End;


var weights: TAofReal;
    refval, resval: string; p,q, it: integer;

begin   //@main
    writeln( 'Target output for the function f(x) = 2x + 1:' );
		Readln;
    showTargetOutput(refval);
    randomWeights(weights);
    writeln('Output from untrained perceptron:' );
    showOutput( weights, resval );
    for it:= 0 to 2 do write(floattostr(weights[it])+', ');
    writeln('--------------------------------------');
    train( weights, 1, LEARNRATE);
    writeln('Output from perceptron after 1 training run:' );
    showOutput( weights, resval );
    for it:= 0 to 2 do write(floattostr(weights[it])+', ');
    writeln('--------------------------------------');
    train( weights, 14, LEARNRATE);
    writeln('Output from perceptron after 14 training runs:' );
    showOutput( weights, resval );
    for it:= 0 to 2 do writeln(format('weights %10.4f ',[weights[it]]));
    writeln('--------------------------------------');
    if compareStr(refval,resval)= 0 then
       writeln('accuracy ~100') else
         writeln('accuracy NOT 100');
    writeln('');
	  writeln('Now testing values-----------');
    testAll( weights, 1 );
    writeln('');
    writeln('Predict values---------------');
    Predict( weights, 15,5);
    Writeln('');
    writeln('Predict line values----------');
    for p:= -9 to 10 do
      //for q:= -9 to 10 do
        Predict(weights, p, p);
    // TEst Float
    for p:= -9 to 10 do
      //for q:= -9 to 10 do
        PredictFloat(weights, p, p);
		sleep(1000);
		//halt(1);
		Readln;
	  Readln;

End.


ref: delta list, 0 is no error

err(0,2)/targ(-1,1)/feed(-1,1)
-----------------------
   0   :    1    1
   0   :   -1   -1
   2   :    1   -1
  -2   :   -1    1

optimal weights by 4000 runs
weights    -1.0114
weights     0.5006
weights    -0.5560
--------------------------------------

ref:

https://maxbox4.wordpress.com/code/
https://www.academia.edu/36608990/TensorFlow_AI_Demo
https://www.scribd.com/document/378905755/tensorflow-machinelearning-task9

----File newtemplate.txt exists - cached - now saved!

